function[Lopt, Ropt, Zopt, infos, options] = fixedrank_matrixcompletion(problem, Uinit, options)
    % Problem description
    d = problem.d;
    T = problem.T;
    r = problem.r;
    C = problem.C;
    I = problem.I; % Row indices
    J = problem.J; % Column indices
    Y = problem.Y; % Known entries
    
    % sorting
    mask = sparse(double(I), double(J), Y, d, T, length(I));
    [I, J, Y] = find(mask);
    I = uint32(I);
    J = uint32(J);
    
    length_entries = length(I); % Number of known entries
    if isfield(problem,'I_test') && ~isempty(problem.I_test)
        I_test = problem.I_test;
        J_test = problem.J_test;
        Y_test = problem.Y_test;
        length_entries_test = length(I_test);
    end
    
    % Some sparse structures that are often used.
    mask = sparse(double(I), double(J), ones(length(I), 1), d, T, length(I));
    Z_sparse_mat = sparse(double(I), double(J), ones(length(I), 1), d, T, length(I));
    
    % Local defaults for options
    localdefaults.maxiter = 50; % Max iterations.
    localdefaults.tolgradnorm = 1e-6; % Absolute tolerance on Gradnorm.
    localdefaults.maxinner = 30; % Max inner iterations for the tCG step.
    localdefaults.tolrelgradnorm = 1e-10; % Gradnorm/initGradnorm tolerance.
    localdefaults.method = 'TR'; % Default solver is trustregions (TR).
    localdefaults.verbosity = 2; % Default: show the output.
    
    % Manifold geometry
    problem.M = spectrahedronfactory(d, r);
    
    % Initialization
    if ~exist('Uinit', 'var')
        Uinit = [];
    end
    
    if ~exist('options', 'var') || isempty(options)
        options = struct();
    end
    options = mergeOptions(localdefaults, options);
    
    
    % Cost
    problem.cost = @cost;
    function [f, store] = cost(U, store)
        if ~isfield(store, 'Z')
            [store.Z, store.UtZ, store.Zchol] = squareLoss_dual_get_Z(U,C,I, J, Y,mask);
        end
        Z = store.Z;
        UtZ = store.UtZ;
        f = (Y'*Z) - (Z'*Z)/(4*C) - 0.5*norm(UtZ,'fro')^2;
    end
    
    % Gradient
    problem.egrad = @egrad;
    function [grad, store] = egrad(U, store)
        if ~isfield(store, 'Z')
            [~, store] = cost(U, store);
        end
        Z = store.Z;
        UtZ = store.UtZ;
        grad = -mymultsparsefull(Z, UtZ', mask); % Equivalent to grad = -(Z*ZtU).
    end
    
    % Hessian
    problem.ehess = @ehess;
    function [graddot, store] = ehess(U, Udot,store)
        if ~isfield(store, 'Z')
            [~, store] = cost(U, store);
        end
        Z = store.Z;
        UtZ = store.UtZ;
        Zchol = store.Zchol;
        
        % Get Zdot at U and Udot.
        [Zdot, UtZdot, UdottZ] = squareLoss_dual_get_Zdot(U,C,I, J,mask,Udot,Z,UtZ,Zchol);
        
        graddot = -mymultsparsefull(Zdot, UtZ', mask)...
            -mymultsparsefull(Z, (UtZdot+UdottZ)', mask);
    end
    
    % Stats that we compute every iteration
    options.statsfun = @mystatsfun;
    function stats = mystatsfun(problem, U, stats, store)
        primal_cost = stats.cost; % Equal to the partial dual cost.
        mysetsparseentries(Z_sparse_mat, store.Z);
        sigmax_Z = svds(Z_sparse_mat, 1); % Maximum singular value of Z.
        dual_cost = (Y'*store.Z) - (store.Z'*store.Z)/(4*C) - 0.5*sigmax_Z^2;
        stats.duality_gap = abs(primal_cost - dual_cost); % Absolute duality gap
        stats.rel_duality_gap = abs(stats.duality_gap/dual_cost); % Relative duality gap
        
        % Primal low-rank factors
        L = U;
        R = store.UtZ;
        
        % Compute train RMSE
        error_train = myspmaskmult(L', R, I, J) ...
            - Y;
        stats.loss = norm(error_train)^2; % Square error on train data.
        stats.loss_rmse = sqrt((error_train'*error_train)/length_entries); % RMSE on train data.
        
        if exist('I_test', 'var')
            error_test = myspmaskmult(L', R, I_test, J_test) ...
                - Y_test;
            stats.loss_test = norm(error_test)^2; % Square error on test data.
            stats.loss_test_rmse = sqrt((error_test'*error_test)/length_entries_test); % RMSE on test data.
            if options.verbosity >1
                fprintf('LossRMSE %e TestLossRMSE %e Rel. duality_gap gap %e\n', stats.loss_rmse, stats.loss_test_rmse, stats.rel_duality_gap);
            end
        end
    end
    
    % Additional stopping criteria
    options.stopfun = @mystopfun;
    function stopnow = mystopfun(problem, U, info, last)
        stopnow = (last >= 3 && info(last).gradnorm/info(1).gradnorm < options.tolrelgradnorm);
    end
    
    
    %     % Gradient and Hessian checks
    %     checkgradient(problem);
    %     pause;
    %     checkhessian(problem);
    %     pause;
    
    % Solver
    if strcmpi('TR', options.method)
        % Riemannian trustregions
        [Uopt,~,infos] = trustregions(problem, Uinit, options);
    elseif strcmpi('CG', options.method)
        % Riemannian conjugategradients
        options.beta_type = 'H-S';
        options.linesearch = @linesearch;
        options.ls_contraction_factor = .2;
        options.ls_optimism = 1.1;
        options.ls_suff_decr = 1e-4;
        options.ls_max_steps = 25;
        [Uopt,~,infos] = conjugategradient(problem, Uinit, options);
    end
    
    % Compute optimal Z
    [Zopt,UopttZ] = squareLoss_dual_get_Z(Uopt,C,I, J, Y,mask);
    
    % Optimal low-rank factors
    Lopt = Uopt;
    Ropt = UopttZ;
    
end
