/*=================================================================
% function R = spbuildmatrix(Ct, U, lambda)
%
% Specific function to reduce the computation time in a bottleneck
% portion of the Matlab code for our low-rank matrix completion
% software.
%
% Ct is a *sparse* m-by-n real matrix, U is a (real)
% m-by-r matrix and lambda is a strictly positive real number.
% R is a cell containing n upper triangular r-by-r matrices: the
% Cholesky factors of the diagonal blocks of A.'*A. See the notes
% accompanying the software for details.
%
% The equivalent Matlab code is the following:
%
%         AtA = cell(n, 1);
%         for i = 1 : n
%             
%             I = find(Ct(:, i));
%             scaledUi = U(I, :);
%             AtAi = scaledUi.'*scaledUi + eye(r) / (2*lambda);
%             
%             % We actually store the Cholesky factorization
%             AtA{i} = chol(AtAi);
%             
%         end
%
% 
% Compile with: mex -lmwlapack -lmwblas -largeArrayDims preprocess_parallel.c  CFLAGS="\$CFLAGS -fopenmp" LDFLAGS="\$LDFLAGS -fopenmp"
%
%
 *=================================================================*/

#include "mex.h"
#include "matrix.h"
#include "lapack.h"
#include <math.h>
#include "blas.h"
#include <string.h>
#include <omp.h>

/* Input Arguments */

#define	Ct	    prhs[0]
#define	Ut	    prhs[1]
#define lambda  prhs[2]
#define UomegatYomega  prhs[3]
#define numThreads prhs[4]

/* Output Arguments */

#define R	plhs[0]
#define X   plhs[1]
#define finalM1   plhs[2]
#define finalM2   plhs[3]

void mexFunction(
          int nlhs, mxArray *plhs[], 
		  int nrhs, const mxArray* prhs[] )
{
    mwSize m, n, r;
    double *Ctvals, *Utvals, *UomegatYomegavals;
    ptrdiff_t N, M, pone;
    double lambdaval, const1, const2;
    mwIndex *CtIr, *CtJc;

    double one = 1.0;
    double zero = 0.0;
    
        
    m = mxGetM(Ct);
    n = mxGetN(Ct);
    r = mxGetM(Ut);
    if(mxGetN(Ut) != m)
        mexErrMsgTxt("Dimensions mismatch in input arguments.");
    
    if(!mxIsSparse(Ct))
        mexErrMsgTxt("Ct ought to be a sparse matrix.");
    
    lambdaval = 1.0 / (2.0 * mxGetScalar(lambda));
    const1 = 2.0*mxGetScalar(lambda);
    const2 = -const1;
    pone = (ptrdiff_t) 1;
    
    CtIr = mxGetIr(Ct);
    CtJc = mxGetJc(Ct);
    Ctvals = mxGetPr(Ct);
    Utvals = mxGetPr(Ut);
    UomegatYomegavals = mxGetPr(UomegatYomega);
    
    N  = (ptrdiff_t) n;
    M  = (ptrdiff_t) m;
    
    /* Create a suitable cell */
    R = mxCreateCellMatrix(n, 1);
    /*mxArray *X;*/
    X = mxCreateCellMatrix(n, 1);
    mxArray *Z;
    Z = mxCreateCellMatrix(n, 1);

    finalM1 = mxCreateCellMatrix(n, 1);
    finalM2 = mxCreateDoubleMatrix(r, n, mxREAL);

    mwIndex ii;

    for(ii = 0; ii < n; ++ii)
    {
        mxSetCell(R, ii, mxCreateDoubleMatrix(r, r, mxREAL));
        mxSetCell(X, ii, mxCreateDoubleMatrix(r, r, mxREAL));
        mxSetCell(finalM1, ii, mxCreateDoubleMatrix(r, r, mxREAL));
        mxSetCell(Z, ii, mxCreateDoubleMatrix(r, r, mxREAL));
    }
    double *finalM2ivals;
    finalM2ivals = mxGetPr(finalM2);

    int maxNrThreads = (int) mxGetScalar(numThreads);
    /* Create a dummy cell with matrices which will be used for memory allocation*/
    mxArray *scaledUt;
    scaledUt = mxCreateCellMatrix(maxNrThreads, 1);
    for(ii = 0; ii < maxNrThreads; ++ii)
    {
        /* allocate space for an r-by-r matrix in each cell entry */
        mxSetCell(scaledUt, ii, mxCreateDoubleMatrix(r, m, mxREAL));
    }


    omp_set_num_threads(maxNrThreads);
    #pragma omp parallel
    {
        mwIndex i;
        int id = omp_get_thread_num();
        int nthrds = omp_get_num_threads();

        char *uplo = "U";
        char *TRANS = "T";
        char *NOTRANS = "N";
        mwSize nb;
        ptrdiff_t NB;
        int j, k, temp1, temp2;
        mxArray *Ri, *Xi, *Zi, *finalM1i;
        double *Rivals, *Xivals, *Zivals, *finalM1ivals, *finalM2ivals_temp;
        double *UomegatYomegavals_temp;
        mwIndex *Ctcol;

        mxArray *scaledUti = mxGetCell(scaledUt, id);
        double *scaledUtivals = mxGetPr(scaledUti);
        for(i = id; i < n; i+=nthrds)
        {
            ptrdiff_t info = 0;
            ptrdiff_t RR = (ptrdiff_t) r;
            /* allocate space for an r-by-r matrix in each cell entry */
            Ri = mxGetCell(R, i);
            Rivals = mxGetPr(Ri);

            Xi = mxGetCell(X, i);
            Xivals = mxGetPr(Xi);

            Zi = mxGetCell(Z, i);
            Zivals = mxGetPr(Zi);

            finalM1i = mxGetCell(finalM1, i);
            finalM1ivals = mxGetPr(finalM1i);

            /* compute the entries of the matrix R{i}
               nb is the number of nonzero entries in the i-th column of Ct
               Ctcol is a pointer to an array of nb elements: the indices of
               rows where a nonzero element exists on column i of Ct. */
            nb = CtJc[i+1]-CtJc[i];
            /* mexPrintf("%d ", nb); */
            Ctcol = CtIr+CtJc[i];
            if(nb > 0)
            {
                    for(j = 0; j < nb; ++j)
                    {
                        temp1 = j*r;
                        temp2 = Ctcol[j]*r;
                        for(k = 0; k < r; ++k)
                        {
                            /* scaledUi(j, k) =  U(Ctcol[j], k) */
                            /*scaledUivals[j+k*m] = Uvals[Ctcol[j]+k*m];*/
                            scaledUtivals[k+temp1] = Utvals[k+temp2];
                        }
                    }
                    /*   mexPrintf("\n"); */
                    /* compute the matrix product of the interesting part of
                       scaledUi with itself (transposed): Ri = scaledUi.'*scaledUi */
                    NB = (ptrdiff_t) nb;
                    dgemm(NOTRANS, TRANS, &RR, &RR, &NB, &one,
                          scaledUtivals, &RR, scaledUtivals, &RR, &zero, Rivals, &RR);
                    /*dgemm(TRANS, NOTRANS, &RR, &RR, &NB, &one,
                          scaledUivals, &M, scaledUivals, &M, &zero, Rivals, &RR);*/
            }
            else
            {
                for(j = 0; j < r*r; ++j)
                    Rivals[j] = 0.0;
            }
            
            memcpy(Xivals, Rivals, r*r*sizeof(double));
            memcpy(Zivals, Rivals, r*r*sizeof(double));

            /* add lambda*eye(r) to Ri */
            for(j = 0; j < r; ++j)
                Rivals[j*(r+1)] += lambdaval;
            
            /* remplace R{i} by its Cholesky factor by calling Lapack's dpotrf. */
            dpotrf(uplo, &RR, Rivals, &RR, &info);
            /* write zeroes on the lower triangular parts of Ri */
            /*for(k = 0; k < r-1; ++k)
            {
                temp3 = r*k;
                for(j = k+1; j < r; ++j)
                    Rivals[j+temp3] = 0.0;
            }*/
            
            if(info < 0)
                mexErrMsgTxt("dpotrf (in spbuildmatrix): invalid argument.");
            if(info > 0)
                mexErrMsgTxt("dpotrf (in spbuildmatrix): a leading minor is not positive definite.");

            dpotrs(uplo, &RR, &RR, Rivals, &RR, Zivals, &RR, &info);
            if(info < 0)
                mexErrMsgTxt("dpotrs: error negative.");
            if(info > 0)
                mexErrMsgTxt("dpotrs: error positive.");

            for(j = 0; j < r; ++j)
                Zivals[j*(r+1)] -= 1.0;

            dgemm(NOTRANS, NOTRANS, &RR, &RR, &RR, &const1,
                      Zivals, &RR, Xivals, &RR, &zero, finalM1ivals, &RR);

            for(j = 0; j < r; ++j)
                finalM1ivals[j*(r+1)] += 1.0;

            finalM2ivals_temp = finalM2ivals + r*i;
            UomegatYomegavals_temp = UomegatYomegavals + r*i;
            dgemv(NOTRANS, &RR, &RR, &const2, Zivals, &RR, UomegatYomegavals_temp, &pone, &zero, finalM2ivals_temp, &pone);
        }
    }
    mxDestroyArray(scaledUt);
    mxDestroyArray(Z);
    
    return;
}


