/* Parallel version of the mex code for solving the inner problem for 
* robust matrix factorization algorithm proposed in 
* Pratik Jawanpuria and Bamdev Mishra
* "A Saddle Point Approach to Structured Low-rank Matrix Learning in Large-scale Applications"
* The draft can be downloaded from arXiv. Its link is: https://arxiv.org/abs/1704.07352.
* 
* The implemented co-ordinate descent algorithm is based on the work: 
* Chia-Hua Ho and Chih-Jen Lin. "Large-scale Linear Support Vector Regression", JMLR 2012. 
* Parts of code is based on LIBLINEAR solver (https://www.csie.ntu.edu.tw/~cjlin/liblinear/). 
* 
* Note that this code uses reentrant version of random number generator, 
* which are GNU extensions and are not portable. Hence, this code currently 
* does not compile and run on OSX. However, non-parallel version of this 
* code, myLiblinear.cc is compatible with OSX. 
* 
* Warning: no check of data consistency is performed.
* Compile with:
* mex -lmwblas -largeArrayDims myLiblinear_parallel_epssvr.cc CFLAGS="\$CFLAGS -std=c99 -fopenmp" LDFLAGS="\$LDFLAGS -fopenmp"
% Code written by Pratik Jawanpuria <pratik.iitb@gmail.com> and Bamdev Mishra <bamdevm@gmail.com>.
 *=================================================================*/

#include <omp.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <ctime>
#include <cstring>
#include <stdarg.h>
#include <locale.h>
#include "matrix.h"
#include "mex.h"
#include "blas.h"

#define INF HUGE_VAL

// #undef GETI
// #define GETI(i) (0)
// To support weights for instances, use GETI(i) (i)



inline double get_wall_time() {
  struct timeval time;
  if (gettimeofday(&time,NULL)){
    return 0;
  }
  return (double)time.tv_sec + (double)time.tv_usec * .000001;
}

inline double get_cpu_time(){
    return (double)clock() / CLOCKS_PER_SEC;
}


void initialize_vectormatrix_zero(double *B, size_t row, size_t col){
    for (size_t i=0;i<col;i++)
        for (size_t j=0;j<row;j++)
            B[i*row+j] = 0.0;
}

void initialize_vectormatrix_zero(int *B, size_t row, size_t col){
    for (size_t i=0;i<col;i++)
        for (size_t j=0;j<row;j++)
            B[i*row+j] = 0;
}

void mexFunction(int nout, mxArray *out[],
/* Output variables */
int nin, const mxArray *in[]) /* Input variables */
{
    // fix random seed to have same results for each run
    // (for cross validation)
    srand(1);

    double *Ut = static_cast<double*> (mxGetData(in[0]));
    int r = (int) mxGetM(in[0]);
    double *Ycolvals = static_cast<double*> (mxGetData(in[1]));
    int numYcol = (int) mxGetM(in[1]);
    size_t *maskIr = mxGetIr(in[2]);
    size_t *maskJc = mxGetJc(in[2]);
    int d = (int) mxGetM(in[2]);
    int T = (int) mxGetN(in[2]);
    double C = mxGetScalar(in[3]);
    double p = mxGetScalar(in[4]);
    double eps = mxGetScalar(in[5]);
    int max_iter = (int) mxGetScalar(in[6]);
    int solver_type = (int) mxGetScalar(in[7]);
    int maxNrThreads = (int) mxGetScalar(in[8]);


    ptrdiff_t ptrdiff_t_1 = (ptrdiff_t) 1;
    ptrdiff_t ptrdiff_t_r = (ptrdiff_t) r;

    out[0] = mxCreateDoubleMatrix(0,0,mxREAL);
    mxSetM(out[0],numYcol);
    mxSetN(out[0],1);
    mxSetPr(out[0], (double *)mxMalloc(sizeof(double)*numYcol));
    //mxSetPr(out[0], (double *)mxCalloc(numYcol,sizeof(double)));
    double *A_ = mxGetPr(out[0]);

    out[1] = mxCreateDoubleMatrix(0,0,mxREAL);
    mxSetM(out[1],r);
    mxSetN(out[1],T);
    mxSetPr(out[1], (double *)mxMalloc(sizeof(double)*r*T));
    double *UtA_ = mxGetPr(out[1]);

    initialize_vectormatrix_zero(A_,numYcol,1);
    initialize_vectormatrix_zero(UtA_,r,T);

    double lambda, upper_bound;

    if(solver_type == 12){
        //L2-svr loss
        lambda = 0.5/C;
        upper_bound = INF;
    }
    else if (solver_type == 13){
        //L1-svr loss
        lambda = 0;
        upper_bound = C;
    }
    else 
        printf("[myLiblinear_parallel_epssvr]ERROR!!! Wrong option provided: %d.\n", solver_type);

    double qii;
    double *QD = new double[d];
    for (int ii=0;ii<d;ii++){
        qii = dnrm2(&ptrdiff_t_r,Ut+ii*r,&ptrdiff_t_1);
        QD[ii] = qii*qii;
    }
    
    omp_set_num_threads(maxNrThreads);
    #pragma omp parallel
    {
        unsigned int id, sparse_row_index;
        int nthrds, i, s, iter, t, l, active_size;
        double Gmax_new, Gnorm1_new, Gmax_old, Gnorm1_init, delta, G, H;
        size_t *maskTcol;
        int *index;
        double *w;
        double *y;
        double *beta;
        double beta_sparse_row_index;

        double Gp, Gn, violation, beta_old;
        int jj;
        int j, temp1, temp2;

        id = omp_get_thread_num();
        nthrds = omp_get_num_threads();
        unsigned int seed;

        index = new int[d];
        for (t=id; t<T; t+=nthrds)
        {
            iter = 0;
            Gmax_old = INF;
            Gnorm1_init = -1.0; // Gnorm1_init is initialized at the first iteration
            l = static_cast<int>(maskJc[t+1] - maskJc[t]);
            active_size = l;
            for(i=0; i<l; i++)
                index[i] = i;
            maskTcol = maskIr+maskJc[t];
            w = UtA_ + r*t;
            beta = A_ + maskJc[t];
            y = Ycolvals + maskJc[t];
            while (iter < max_iter)
            {
                Gmax_new = 0;
                Gnorm1_new = 0;
                for(i=0; i<active_size; i++)
                {
                    jj = rand_r(&seed);
            		j = i + (int)(jj%(active_size-i));
                    temp1 = index[i];
                    index[i] = index[j];//temp2;
                    index[j] = temp1;
                }
                for(s=0; s<active_size; s++)
                {
                    sparse_row_index = index[s];
                    i = maskTcol[sparse_row_index];
                    beta_sparse_row_index = beta[sparse_row_index];
                    G = -y[sparse_row_index] + lambda*beta_sparse_row_index;
                    H = QD[i] + lambda;

                    G += ddot(&ptrdiff_t_r,Ut+i*r,&ptrdiff_t_1,w,&ptrdiff_t_1);

                    Gp = G+p;
                    Gn = G-p;
                    violation = 0;
                    if(beta_sparse_row_index == 0)
                    {
                        if(Gp < 0)
                            violation = -Gp;
                        else if(Gn > 0)
                            violation = Gn;
                        else if(Gp>Gmax_old && Gn<-Gmax_old)
                        {
                            active_size--;
                            temp1 = index[s];
                            index[s] = index[active_size];//temp2;
                            index[active_size] = temp1;
                            s--;
                            continue;
                        }
                    }
                    else if(beta_sparse_row_index >= upper_bound)
                    {
                        if(Gp > 0)
                            violation = Gp;
                        else if(Gp < -Gmax_old)
                        {
                            active_size--;
                            temp1 = index[s];
                            index[s] = index[active_size];//temp2;
                            index[active_size] = temp1;
                            s--;
                            continue;
                        }
                    }
                    else if(beta_sparse_row_index <= -upper_bound)
                    {
                        if(Gn < 0)
                            violation = -Gn;
                        else if(Gn > Gmax_old)
                        {
                            active_size--;
                            temp1 = index[s];
                            index[s] = index[active_size];//temp2;
                            index[active_size] = temp1;
                            s--;
                            continue;
                        }
                    }
                    else if(beta_sparse_row_index > 0)
                        violation = fabs(Gp);
                    else
                        violation = fabs(Gn);

                    Gmax_new = fmax(Gmax_new, violation);
                    Gnorm1_new += violation;

                    // obtain Newton direction d
                    if(Gp < H*beta_sparse_row_index)
                        delta = -Gp/H;
                    else if(Gn > H*beta_sparse_row_index)
                        delta = -Gn/H;
                    else
                        delta = -beta_sparse_row_index;

                    if(fabs(delta) < 1.0e-12)
                        continue;

                    beta_old = beta_sparse_row_index;
                    beta[sparse_row_index] = fmin(fmax(beta_sparse_row_index+delta, -upper_bound), upper_bound);
                    delta = beta[sparse_row_index]-beta_old;

                    if(delta != 0)
                        daxpy(&ptrdiff_t_r, &delta, Ut+i*r, &ptrdiff_t_1, w, &ptrdiff_t_1);
                }
                if(iter == 0)
                    Gnorm1_init = Gnorm1_new;
                iter++;
                // if(iter % 10 == 0)
                //     info(".");

                if(Gnorm1_new <= eps*Gnorm1_init)
                {
                    if(active_size == l){
                        break;
                    }
                    else
                    {
                        active_size = l;
                        // info("*");
                        Gmax_old = INF;
                        continue;
                    }
                }
                Gmax_old = Gmax_new;
            }
        }
        delete[] index;
    }
    delete[] QD;
}




