/*=================================================================
% function X = cholsolve(R, B)
% Computes X such that A{i}X(:,i) = B(:,i), where R{i} is the upper
% triangular Cholesky factor of the real symmetric positive-definite
% matrix A{i}, i.e., A{i} = R{i}.'*R{i}.
% R is a cell array of m upper cholesky factors of real, symmetric,
% positive-definite matrices A{i}, i=1..m, of size n-by-n;
% B and X are n-by-m real matrices.
%
% Compile with: mex -lmwlapack -largeArrayDims mycholsolvecell_parallel.c CFLAGS="\$CFLAGS -fopenmp" LDFLAGS="\$LDFLAGS -fopenmp"
%
% Code originally written by Nicolas Boumal <nicolasboumal@gmail.com>.
% Original code has been modified by Pratik Jawanpuria <pratik.iitb@gmail.com> and Bamdev Mishra <bamdevm@gmail.com>.
 *=================================================================*/

#include <omp.h>
#include "mex.h"
#include "string.h"
#include "matrix.h"
#include "lapack.h"
#include <math.h>
#include <sys/time.h>

/* Timing functions */
struct timeval start()
{
    struct timeval tm ;
    gettimeofday(&tm, NULL);
    return tm ;
}

void stop(struct timeval tm1)
{
    struct timeval tm2;
    gettimeofday(&tm2, NULL);
    unsigned long long t = 1000 * (tm2.tv_sec - tm1.tv_sec) + (tm2.tv_usec - tm1.tv_usec) / 1000;
    printf("%llu ms\n", t);
}

unsigned long long stopAndGet(struct timeval tm1)
{
    struct timeval tm2;
    gettimeofday(&tm2, NULL);
    unsigned long long t = 1000 * (tm2.tv_sec - tm1.tv_sec) + (tm2.tv_usec - tm1.tv_usec) / 1000;
    return t ;
}

double stopAndGetDouble(struct timeval tm1)
{
    return (double)stopAndGet(tm1) ;
}

unsigned long long getMeanTimes(unsigned long long* times, int* nOfIterations, int maxthreads)
{
    unsigned long long meanTime = 0 ;
    int i = 0 ;
    for(i = 0 ; i < maxthreads ; i++)
    {
        if(nOfIterations[i] > 0)
            meanTime += times[i] ; /* /nOfIterations[i] ;   */
    }
    return meanTime ;
}


/* Input Arguments */

#define	R	prhs[0]
#define	B	prhs[1]
#define numThreads prhs[2]

/* Output Arguments */

#define	X	plhs[0]

void mexFunction(
          int nlhs, mxArray *plhs[], 
		  int nrhs, const mxArray* prhs[] )
{
    mwIndex k1, k2; /* debug */
    /*double *b;*/
    
    mwIndex i;
    mwSize m, n;
    /*mxArray *Ri;*/
    double *Bvals, *Xvals;
    char *uplo = "U";

    /* !! very important to use the proper typedef since confusion between
       32 and 64 bits for integers is quite common. */
    ptrdiff_t info = 0;
    ptrdiff_t one = 1;
	ptrdiff_t N = 0;
    
    /* Check for proper number of arguments */
    if (nrhs != 3) { 
        mexErrMsgTxt("Three input arguments required."); 
    } else if (nlhs != 1) {
        mexErrMsgTxt("Single output argument required."); 
    }
    
    if(!mxIsCell(R))
        mexErrMsgTxt("CHOLSOLVECELL: R must be a cell containing m upper triangular matrices of size n-be-n.");
        
    /* retrieve dimensions of the matrices */
    n = mxGetM(B);
    m = mxGetN(B);
    if(m <= 0 || n <= 0 || mxGetNumberOfElements(R) != m)
        mexErrMsgTxt("CHOLSOLVECELL: inconsistent matrix / cell dimensions.");
    
    /* dpotrs will store the result in the B matrix we pass to it;
       hence we copy B in a new matrix X and pass X to dpotrs.
       As a result, B will be unaltered and dpotrs will write the
       system solutions in the output matrix directly. */
    X = mxCreateDoubleMatrix(n, m, mxREAL);
    Xvals = mxGetPr(X);
    Bvals = mxGetPr(B);
    memcpy(Xvals, Bvals, m*n*sizeof(double));
    
    /* Call LAPACK subroutine (this is where the magic happens) */
	N = (ptrdiff_t) n;
    int maxNrThreads = (int) mxGetScalar(numThreads); /*16*/
    omp_set_num_threads(maxNrThreads);
    /*int nn;
     #pragma omp for
     for(nn=0; nn<10; ++nn)
     {
       printf(" %d", nn);
     }
     printf(".\n");*/

    /*struct timeval start_forW = start() ;*/
    /*printf("%d\n",m);*/
    #pragma omp parallel for schedule(static,1000) /*private(i,Ri,Rvals,b) schedule(dynamic,1000)*/
    for(i = 0; i < m; ++i)
    {   
        /*int tidLoc = omp_get_thread_num();
        printf("Thread: %d\n",tidLoc);*/
        double *b;
        mxArray *Ri;
        double *Rvals;
        Ri = mxGetCell(R, i);
        /*if(mxGetM(Ri) != n || mxGetN(Ri) != n)
            mexErrMsgTxt("CHOLSOLVECELL: inconsistent matrix / cell dimensions.");*/
        Rvals = mxGetPr(Ri);
        b = Xvals+n*i;
        dpotrs(uplo, &N, &one, Rvals, &N, b, &N, &info);
        /*if(info < 0)
            mexErrMsgTxt("dpotrs (in cholsolvecell): invalid argument.");*/
    }
    /*mexPrintf("Time: ") ;
    stop(start_forW) ;*/
    
    return;
}


